#  Copyright (c) 2020. Robin Thibaut, Ghent University
import warnings
from os.path import join as jp

import matplotlib.pyplot as plt
import numpy as np
from loguru import logger

from pysgems.base.packbase import Package
from pysgems.io.sgio import datread


class Plots(Package):
    def __init__(self, project):
        Package.__init__(self, project)
        self.name = self.parent.project_name

    def plot_coordinates(self):
        try:
            # Scatter plot (2D)
            plt.plot(
                self.parent.point_set.raw_data[:, 0],
                self.parent.point_set.raw_data[:, 1],
                "ko",
            )
        except Exception as e:
            logger.info(e)
            pass
        try:
            plt.xticks(
                np.cumsum(self.parent.dis.along_r)
                + self.parent.dis.xo
                - self.parent.dis.dx,
                labels=[],
            )
            plt.yticks(
                np.cumsum(self.parent.dis.along_c)
                + self.parent.dis.yo
                - self.parent.dis.dy,
                labels=[],
            )
        except Exception as e:
            logger.info(e)
            pass

        plt.grid("blue")
        plt.show()

    def plot_2d(self, name, res_file=None, save=False):
        """Rudimentary 2D plot

        :param name: name of the plot
        :type name: str
        :param res_file: name of the file to plot
        :param save: save the plot
        """
        if res_file is None:
            res_file = jp(self.parent.res_dir, "results.grid")
        matrix = datread(res_file, start=3)
        if matrix.ndim > 2:
            warnings.warn("Only 2D matrices supported, dimensionality exceeded")
            return 0
        matrix = np.where(matrix == -9966699, np.nan, matrix)
        matrix = matrix.reshape((self.parent.dis.nrow, self.parent.dis.ncol))
        extent = (
            self.parent.dis.xo,
            self.parent.dis.x_lim,
            self.parent.dis.yo,
            self.parent.dis.y_lim,
        )
        plt.imshow(np.flipud(matrix), cmap="coolwarm", extent=extent)
        plt.plot(
            self.parent.point_set.raw_data[:, 0],
            self.parent.point_set.raw_data[:, 1],
            "k+",
            markersize=1,
            alpha=0.7,
        )
        plt.colorbar()
        if save:
            plt.savefig(jp(self.parent.res_dir, name), bbox_inches="tight", dpi=300)
        plt.show()



#  Copyright (c) 2020. Robin Thibaut, Ghent University

import os
import shutil
import time
from os.path import join as jp

import numpy as np
import pandas as pd
from loguru import logger

from pysgems.base.packbase import Package


def blocks_from_rc(
    rows: np.array,
    columns: np.array,
    layers: np.array,
    xo: float = 0,
    yo: float = 0,
    zo: float = 0,
):
    """
    Yields blocks defining grid cells

    :param rows: array of x-widths along a row
    :param columns: array of y-widths along a column
    :param layers: array of z-widths along a column
    :param xo: x origin
    :param yo: y origin
    :param zo: z origin
    :return: generator of (cell node number, block vertices coordinates, block center)
    """
    nrow = len(rows)
    ncol = len(columns)
    nlay = len(layers)
    delr = rows
    delc = columns
    dell = layers
    r_sum = np.cumsum(delr) + yo
    c_sum = np.cumsum(delc) + xo
    l_sum = np.cumsum(dell) + zo

    def get_node(r: int, c: int, h: int) -> int:
        """
        Get node index to fix hard data

        :param r: row number
        :param c: column number
        :param h: layer number
        :return: node number
        """
        nrc = nrow * ncol
        return int((h * nrc) + (r * ncol) + c)

    for k in range(nlay):
        for i in range(nrow):
            for j in range(ncol):
                b = [
                    [c_sum[j] - delc[j], r_sum[i] - delr[i], l_sum[k] - dell[k]],
                    [c_sum[j], r_sum[i] - delr[i], l_sum[k] - dell[k]],
                    [c_sum[j] - delc[j], r_sum[i], l_sum[k] - dell[k]],
                    [c_sum[j], r_sum[i], l_sum[k] - dell[k]],
                    [c_sum[j] - delc[j], r_sum[i] - delr[i], l_sum[k]],
                    [c_sum[j], r_sum[i] - delr[i], l_sum[k]],
                    [c_sum[j] - delc[j], r_sum[i], l_sum[k]],
                    [c_sum[j], r_sum[i], l_sum[k]],
                ]
                yield get_node(i, j, k), np.array(b), np.mean(b, axis=0)


class Discretize(Package):
    def __init__(
        self,
        project,
        dx: float = 1,
        dy: float = 1,
        dz: float = 1,
        xo: float = None,
        yo: float = None,
        zo: float = None,
        x_lim: float = None,
        y_lim: float = None,
        z_lim: float = None,
    ):
        """
        Constructs the grid geometry. The user can not control directly the number of rows and columns
        but instead inputs the cell size in x and y dimensions.
        xo, yo, x_lim, y_lim, defining the bounding box of the grid, are None by default, and are computed
        based on the data points distribution.

        :param project: sgems project
        :param dx: cell size in x dimension
        :param dy: cell size in y dimension
        :param dz: cell size in z dimension
        :param xo: x origin
        :param yo: y origin
        :param zo: z origin
        :param x_lim: x limit
        :param y_lim: y limit
        :param z_lim: z limit
        """

        Package.__init__(self, project)

        self.cell_file = None

        self.dx = dx
        self.dy = dy
        self.dz = dz
        if self.parent.point_set is not None:
            if self.parent.point_set.dimension == 2:
                self.dz = 0

        # Grid origins
        # If a PointSet is already loaded into the project, automatically defines the limits
        # (if no limits are given)
        if xo is None:
            if self.parent.point_set is not None:
                xs = self.parent.point_set.dataframe["x"]
                xo = np.min(xs) - 4 * self.dx
            else:
                xo = 0

        if yo is None:
            if self.parent.point_set is not None:
                ys = self.parent.point_set.dataframe["y"]
                yo = np.min(ys) - 4 * self.dy
            else:
                yo = 0

        if zo is None:
            if self.parent.point_set is not None:
                zs = self.parent.point_set.dataframe["z"]
                zo = np.min(zs) - 4 * self.dz
            else:
                zo = 0

        # Grid limits
        if x_lim is None:
            if self.parent.point_set is not None:
                xs = self.parent.point_set.dataframe["x"]
                x_lim = np.max(xs) + 4 * self.dx
            else:
                x_lim = 1

        if y_lim is None:
            if self.parent.point_set is not None:
                ys = self.parent.point_set.dataframe["y"]
                y_lim = np.max(ys) + 4 * self.dy
            else:
                y_lim = 1

        if z_lim is None:
            if self.parent.point_set is not None:
                zs = self.parent.point_set.dataframe["z"]
                z_lim = np.max(zs) + 4 * self.dz
            else:
                z_lim = 1

        # Cell dimensions
        if self.dy > 0:
            nrow = abs(int((y_lim - yo) // self.dy))  # Number of rows
        else:
            nrow = 1
        if self.dx > 0:
            ncol = abs(int((x_lim - xo) // self.dx))  # Number of columns
        else:
            ncol = 1
        if self.dz > 0:
            nlay = abs(int((z_lim - zo) // self.dz))  # Number of layers
        else:
            nlay = 1

        # Size of each cell along y-dimension - rows
        along_r = np.ones(ncol) * self.dx * np.sign(x_lim)
        # Size of each cell along x-dimension - columns
        along_c = np.ones(nrow) * self.dy * np.sign(y_lim)
        # Size of each cell along x-dimension - columns
        along_l = np.ones(nlay) * self.dz * np.sign(z_lim)

        self.xo = xo
        self.yo = yo
        self.zo = zo

        self.x_lim = x_lim
        self.y_lim = y_lim
        self.z_lim = z_lim

        self.nrow = nrow
        self.ncol = ncol
        self.nlay = nlay

        self.along_r = along_r
        self.along_c = along_c
        self.along_l = along_l

        setattr(self.parent, "dis", self)

    def my_cell(self, xyz: np.array):
        """
        Given a point coordinate xyz [x, y, z], computes its cell number by computing the euclidean distance of each cell
        center.

        :param xyz:  x, y, z coordinate of data point
        :return: cell number
        """
        start = time.time()
        rn = np.array(xyz)
        # first check if point is within the grid

        # crit = 0
        # if np.all(rn >= np.array([self.xo, self.yo, self.zo])) and np.all(
        #     rn <= np.array([self.x_lim, self.y_lim, self.z_lim])
        # ):
        #     crit = 1

        crit = 1
        if crit:  # if point inside
            if self.parent.point_set.dimension == 3:  # if 3D
                # minimum distance under which a point is in a cell
                dmin = np.min([self.dx, self.dy, self.dz]) / 2
            else:  # if 2D
                # minimum distance under which a point is in a cell
                dmin = np.min([self.dx, self.dy]) / 2

            blocks = blocks_from_rc(
                self.along_c, self.along_r, self.along_l, self.xo, self.yo, self.zo
            )  # cell generator

            # mapping data points to cells:
            # slow but memory-effective method
            vmin = np.inf
            cell = None
            for b in blocks:
                c = b[2]
                dc = np.linalg.norm(rn - c)  # Euclidean distance
                if dc <= dmin:  # If point is inside cell
                    print("found 1 cell id in {} s".format(time.time() - start))
                    return b[0]
                if dc < vmin:
                    vmin = dc
                    cell = b[0]
            print("found 1 cell id in {} s".format(time.time() - start))
            return cell
        else:
            return self.parent.nodata

    def compute_cells(self, xyz: np.array):
        """
        Determines cell location for each data point.
        It is necessary to know the cell number to assign the hard data property to the sgems grid.

        :param xyz: Data points x, y, z coordinates
        """

        nodes = np.array([self.my_cell(c) for c in xyz])

        # Save to nodes to avoid recomputing each time
        np.save(self.cell_file, nodes)

    def write_hard_data(
        self,
        subdata: pd.DataFrame,
        dis_file: str = None,
        cell_file: str = None,
        output_dir: str = None,
    ):
        """
        Removes no-data rows from data frame and compute the mean of data points sharing the same cell.
        Export the list of shape (n features, m cells, 2) containing the node of each point data with the corresponding
        value, for each feature.

        :param dis_file: path to the dis file
        :param cell_file: path to the cell file
        :param subdata: Pandas dataframe whose columns are the values of features to save as hard data.
        :param output_dir: Directory where hard data lists will be saved
        :return: Filtered list of each attribute
        """

        if cell_file is not None:
            self.cell_file = cell_file

        if self.cell_file is None:
            self.cell_file = jp(self.parent.res_dir, "cells.npy")

        if output_dir is None:
            output_dir = self.parent.res_dir

        if dis_file is None:
            dis_file = jp(os.path.dirname(self.cell_file), "grid.dis")

        xyz = subdata[["x", "y", "z"]].to_numpy()

        npar = np.array(
            [
                self.dx,
                self.dy,
                self.dz,
                self.xo,
                self.yo,
                self.zo,
                self.x_lim,
                self.y_lim,
                self.z_lim,
                self.nrow,
                self.ncol,
                self.nlay,
            ]
        )

        if os.path.isfile(dis_file):  # Check previous grid info
            pdis = np.loadtxt(dis_file)
            # If different, recompute data points cell by deleting previous cell file
            if not np.array_equal(pdis, npar):
                logger.info("New grid found")
                try:
                    os.remove(self.cell_file)
                except FileNotFoundError:
                    pass
                finally:
                    np.savetxt(dis_file, npar)
                    self.compute_cells(xyz)
            else:
                logger.info("Using previous grid")
                if not os.path.exists(self.cell_file):
                    self.compute_cells(xyz)
        else:
            np.savetxt(dis_file, npar)
            self.compute_cells(xyz)

        data_nodes = np.load(self.cell_file)
        unique_nodes = list(set(data_nodes))

        h = subdata.columns.values[-1]
        hd = []
        # fixed nodes = [[node i, value i]....]
        fixed_nodes = np.array(
            [[data_nodes[dn], subdata[h][dn]] for dn in range(len(data_nodes))]
        )
        # Deletes points where val == nodata
        hard_data = np.delete(
            fixed_nodes, np.where(fixed_nodes == self.parent.nodata)[0], axis=0
        )
        # If data points share the same cell, compute their mean and assign the value to the cell
        for n in unique_nodes:
            where = np.where(hard_data[:, 0] == n)[0]
            if len(where) > 1:  # If more than 1 point per cell
                mean = np.mean(hard_data[where, 1])
                hard_data[where, 1] = mean

            fn = hard_data.tolist()
            # For each, feature X, saves a file X.hard
            cell_values_name = jp(os.path.dirname(self.cell_file), f"{h}.hard")
            with open(cell_values_name, "w") as nd:
                nd.write(repr(fn))
            try:
                shutil.copyfile(
                    cell_values_name,
                    cell_values_name.replace(
                        os.path.dirname(cell_values_name), output_dir
                    ),
                )
            except shutil.SameFileError:
                pass

        setattr(self.parent, "hard_data", hd.append(h))



import os

import sgems

nodata = -9966699

os.chdir("RES_DIR")
sgems.execute("DeleteObjects computation_grid")
sgems.execute("DeleteObjects PROJECT_NAME")
sgems.execute("DeleteObjects finished")

for file in OBJECT_FILES:
    sgems.execute("LoadObjectFromFile  {}::All".format(file))

sgems.execute("NewCartesianGrid  computation_grid::GRID")

#~sgems.execute('RunGeostatAlgorithm  ALGORITHM_NAME::/GeostatParamUtils/XML::ALGORITHM_XML')
#~sgems.execute('SaveGeostatGrid  computation_grid::FEATURE_OUTPUT.grid::gslib::0::OUTPUT_LIST')
if "kriging" in "ALGORITHM_NAME":  # Save variance grid
#~    sgems.execute('SaveGeostatGrid  computation_grid::FEATURE_OUTPUT_var.grid::gslib::1::OUTPUT_LIST_krig_var')


#  Copyright (c) 2020. Robin Thibaut, Ghent University


def joinlist(j, mylist):
    """
    Function that joins an array of numbers with j as separator. For example, joinlist('^', [1,2]) returns 1^2

    :param j: separator
    :param mylist: list of numbers
    """
    gp = j.join(map(str, mylist))

    return gp



#  Copyright (c) 2020. Robin Thibaut, Ghent University

import os
from os.path import join as join_path

from pysgems.algo.sgalgo import XML
from pysgems.dis.sgdis import Discretize
from pysgems.io.sgio import PointSet
from pysgems.plot.sgplots import Plots
from pysgems.sgems import sg


def main():
    # %% Initiate sgems pjt
    cwd = os.getcwd()  # Working directory
    rdir = join_path(cwd, "results", "demo_kriging")  # Results directory
    pjt = sg.Sgems(project_name="sgems_test", project_wd=cwd, res_dir=rdir)

    # %% Load data point set
    data_dir = join_path(cwd, "datasets", "demo_kriging")
    dataset = "sgems_dataset.eas"
    file_path = join_path(data_dir, dataset)

    ps = PointSet(project=pjt, pointset_path=file_path)

    # %% Generate grid. Grid dimensions can automatically be generated based on the data points
    # unless specified otherwise, but cell dimensions dx, dy, (dz) must be specified
    ds = Discretize(project=pjt, dx=5, dy=5)

    # %% Display point coordinates and grid
    pl = Plots(project=pjt)
    pl.plot_coordinates()

    # %% Which feature are available
    print(pjt.point_set.columns)

    # %% Load your algorithm xml file in the 'algorithms' folder.
    algo_dir = join_path(os.path.dirname(cwd), "algorithms")
    al = XML(project=pjt, algo_dir=algo_dir)
    al.xml_reader("kriging")

    # %% Show xml structure tree
    al.show_tree()

    # %% Modify xml below:
    # By default, the feature grid name of feature X is called 'X_grid'.
    # sgems.xml_update(path, attribute, new value)
    al.xml_update("Hard_Data", "grid", "ag_grid")
    al.xml_update("Hard_Data", "property", "ag")

    # %% Write binary datasets of needed features
    # sgems.export_01(['f1', 'f2'...'fn'])
    ps.export_01("ag")

    # %% Write python script
    pjt.write_command()

    # %% Run sgems
    pjt.run()

    # Plot 2D results
    name_kriging = "results"
    result_file_kriging = join_path(rdir, f"{name_kriging}.grid")
    save = "kriging"
    pl.plot_2d(name_kriging, res_file=result_file_kriging, save=save)

    name_krigingvar = "results_var"
    result_file_kriging_variance = join_path(rdir, f"{name_krigingvar}.grid")
    save = "kriging_var"
    pl.plot_2d(name_krigingvar, res_file=result_file_kriging_variance, save=save)


if __name__ == "__main__":
    main()



#  Copyright (c) 2020. Robin Thibaut, Ghent University

import abc


class PackageInterface(object):
    @property
    @abc.abstractmethod
    def parent(self):
        raise NotImplementedError(
            "must define get_model_dim_arrays in child " "class to use this base class"
        )

    @parent.setter
    @abc.abstractmethod
    def parent(self, name):
        raise NotImplementedError(
            "must define get_model_dim_arrays in child " "class to use this base class"
        )


class Package(PackageInterface):
    """
    Base package class from which most other packages are derived.

    """

    def __init__(self, parent):
        """
        Package init

        """
        # To be able to access the parent project object's attributes
        self.parent = parent

        return

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, parent):
        self._parent = parent



from setuptools import find_packages, setup

my_pckg = find_packages(include=["pysgems"],
                        exclude=["pysgems.deprecated", "pysgems.develop"])

setup(
    name="pysgems",
    version="1.1.14",
    packages=my_pckg,
    include_package_data=True,
    url="https://github.com/robinthibaut/pysgems",
    license="MIT",
    author="Robin Thibaut",
    author_email="robin.thibaut@UGent.be",
    description="Use SGeMS (Stanford Geostatistical Modeling Software) within Python.",
    long_description="Use SGeMS (Stanford Geostatistical Modeling Software) within Python.",
    install_requires=["numpy", "pandas", "scipy", "matplotlib", "loguru"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



#  Copyright (c) 2020. Robin Thibaut, Ghent University
import os
import subprocess
import time
import uuid
import warnings
from os.path import join as jp

from loguru import logger

from pysgems.utils.sgutils import joinlist


class Sgems:
    def __init__(
        self,
        project_name: str = "sgems_test",
        project_wd: str = "",
        res_dir: str = "",
        script_dir: str = "",
        exe_name: str = "",
        nodata: int = -9966699,  # sgems default value, do not change this
        check_env: bool = True,
        verbose: bool = True,
    ):
        """
        Initialize sgems object.

        :param project_name: Name of the project.
        :param project_wd: Project working directory.
        :param res_dir: Results directory.
        :param script_dir: Script directory.
        :param exe_name: Name of the sgems executable.
        :param nodata: No data value.
        :param check_env: Check if sgems is installed.
        :param verbose: Verbose mode.
        """

        self.verbose = verbose

        logger.add(jp(project_wd, f"{project_name}.log"), rotation="100 MB")

        if self.verbose:
            logger.info(f"Project {project_name} initiated")

        if check_env:
            # First check if sgems installation files are in the user environment variables
            gstl_home = os.environ.get("GSTLAPPLIHOME")
            if not gstl_home:
                msg = "GSTLAPPLIHOME environment variable does not exist"
                warnings.warn(msg)
                logger.warning(msg)
            else:
                msg = "GSTLAPPLIHOME environment variable found"
                logger.info(msg)
                path = os.getenv("Path")
                if gstl_home not in path:
                    msg = f"Variable {gstl_home} does not exist in Path environment variable"
                    warnings.warn(msg)
                    logger.warning(msg)
                if not exe_name:  # If no sgems exe file name is provided,
                    # checks for sgems exe file in the GSTLAPPLIHOME path
                    for file in os.listdir(gstl_home):
                        if (
                            file.endswith(".exe")
                            and ("sgems" in file)
                            and ("uninstall" not in file)
                        ):
                            exe_name = file
                msg = f"sgems exe file : {exe_name} in {gstl_home}"
                logger.info(msg)

        # Project name
        self.project_name = project_name

        # Working directory
        self.project_wd = project_wd
        if not self.project_wd:
            self.project_wd = os.getcwd()

        # Results directory
        self.res_dir = res_dir
        # result directory generated according to project and algorithm name
        if self.res_dir is None:
            # Generate result directory if none is given
            self.res_dir = jp(
                self.project_wd,
                "results",
                "_".join([self.project_name, uuid.uuid1().hex]),
            )
        if not os.path.exists(self.res_dir):
            os.makedirs(self.res_dir)

        # Exe name
        self.exe_name = exe_name

        self.dis = None  # Discretization instance
        self.point_set = None  # Point set manager instance
        self.algo = None  # XML manipulation instance
        self.nodata = nodata

        self.object_file_names = []  # List of features name needed for the algorithm
        self.command_name = ""

        if not script_dir:
            dir_path = os.path.abspath(__file__ + "/../../")
            # Python template file path
            self.template_file = jp(dir_path, "script_templates", "script_template.py")

    def write_command(self):
        """
        Write python script that sgems will run.
        """

        self.command_name = jp(self.res_dir, f"{self.project_name}_commands.py")

        # This empty str will replace the # in front of the commands meant to execute sgems
        run_algo_flag = ""
        # within its python environment
        try:
            name = self.algo.root.find("algorithm").attrib["name"]  # Algorithm name
            try:
                # When performing simulations, sgems automatically add '__realn'
                # to the name of the nth generated property.
                nr = int(self.algo.root.find("Nb_Realizations").attrib["value"])
                name_op = "::".join([name + "__real" + str(i) for i in range(nr)])
            except AttributeError:
                name_op = name

            with open(self.algo.op_file) as alx:  # Remove unwanted \n
                algo_xml = alx.read().strip("\n")

        except AttributeError or FileNotFoundError:
            name = "None"
            name_op = name
            algo_xml = "None"
            run_algo_flag = "#"  # If no algorithm loaded, then just loads the data

        sgrid = [
            self.dis.ncol,
            self.dis.nrow,
            self.dis.nlay,
            self.dis.dx,
            self.dis.dy,
            self.dis.dz,
            self.dis.xo,
            self.dis.yo,
            self.dis.zo,
        ]  # Grid information
        grid = joinlist("::", sgrid)  # Grid in sgems format

        sgems_files = [f"{sf}.sgems" for sf in self.object_file_names]

        # The list below is the list of flags that will be replaced in the sgems python script
        # TODO: add option to change output file name (now default 'results.grid')
        params = [
            [run_algo_flag, "#~"],
            # for sgems convention...
            [self.res_dir.replace("\\", "//"), "RES_DIR"],
            [grid, "GRID"],
            [self.project_name, "PROJECT_NAME"],
            ["results", "FEATURE_OUTPUT"],  # results.grid = output file
            [name, "ALGORITHM_NAME"],
            [name_op, "OUTPUT_LIST"],
            [algo_xml, "ALGORITHM_XML"],
            [str(sgems_files), "OBJECT_FILES"],
        ]

        with open(self.template_file) as sst:
            template = sst.read()
        for i in range(len(params)):  # Replaces the parameters
            template = template.replace(params[i][1], params[i][0])

        with open(self.command_name, "w") as sstw:  # Write sgems python file
            sstw.write(template)

    def script_file(self):
        """Create script file"""
        run_script = jp(self.res_dir, "sgems.script")
        rscpt = open(run_script, "w")
        rscpt.write(" ".join(["RunScript", self.command_name]))
        rscpt.close()

    def bat_file(self):
        """Create bat file"""
        if not os.path.isfile(jp(self.res_dir, "sgems.script")):
            self.script_file()

        batch = jp(self.res_dir, "RunSgems.bat")
        bat = open(batch, "w")
        bat.write(" ".join(["cd", self.res_dir, "\n"]))
        bat.write(" ".join([self.exe_name, "sgems.script"]))
        bat.close()

    def run(self):
        """Call bat file, run sgems"""
        batch = jp(self.res_dir, "RunSgems.bat")
        if not os.path.isfile(batch):
            self.bat_file()
        start = time.time()

        try:
            os.remove(self.algo.op_file)
        except FileNotFoundError:
            pass

        subprocess.call([batch])  # Opens the BAT file
        if self.verbose:
            logger.info(f"ran algorithm in {time.time() - start} s")
